"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ensureBuilderUniqueOrder = exports.polymorphicCondition = exports.polySqlKeyMatch = exports.polyForeignKeyUnique = exports.getPrimaryKey = exports.getPrimaryKeys = exports.generateFieldWithHookFunc = exports.columnToPolyConstraint = exports.isPolymorphicColumn = exports.getSourceColumns = exports.validatePrerequisit = exports.canonical = void 0;
function canonical(str) {
    const m = str.match(/\w+$/);
    return (m && m[0]) || str;
}
exports.canonical = canonical;
function notNull(obj) {
    return !!obj;
}
exports.validatePrerequisit = (build) => {
    const { pgPolymorphicClassAndTargetModels, mapFieldToPgTable, } = build;
    if (!Array.isArray(pgPolymorphicClassAndTargetModels) || !mapFieldToPgTable) {
        throw new Error(`The pgPolymorphicClassAndTargetModels or mapFieldToPgTable is not defined,
    you need to use addModelTableMappingPlugin and definePolymorphicCustom before this`);
    }
};
exports.getSourceColumns = (c) => {
    const { name } = c;
    return {
        sourceTableId: `${name}_id`,
        sourceTableType: `${name}_type`,
    };
};
exports.isPolymorphicColumn = (attr) => {
    return ['r', 'v', 'm'].includes(attr.class.classKind)
        && attr.name.endsWith('_type') && !!attr.tags.isPolymorphic;
};
exports.columnToPolyConstraint = (build, attr) => {
    const { name, tags: { polymorphicTo = [], isPolymorphic }, } = attr;
    const { mapFieldToPgTable, pgIntrospectionResultsByKind: { classById }, } = build;
    let targetTables = [];
    if (typeof polymorphicTo === 'string') {
        targetTables = [canonical(polymorphicTo)];
    }
    else if (Array.isArray(polymorphicTo)) {
        targetTables = polymorphicTo.map(canonical);
    }
    targetTables = Array.from(new Set(targetTables));
    const polymorphicKey = name.substring(0, name.length - 5);
    const newPolyConstraint = {
        name: polymorphicKey,
        from: attr.class,
        to: targetTables.map((mName) => {
            const pgTableSimple = mapFieldToPgTable[mName];
            if (!pgTableSimple)
                return null;
            const c = classById[pgTableSimple.id];
            // Also this class need to have a single column primary key
            const keys = exports.getPrimaryKeys(c);
            if (keys.length !== 1) {
                return null;
            }
            return {
                pgClass: c,
                pKey: keys[0],
                name: mName,
            };
        }).filter(notNull),
    };
    if (typeof isPolymorphic === 'string') {
        // There is a backward association name for this
        newPolyConstraint.backwardAssociationName = isPolymorphic;
    }
    return newPolyConstraint;
};
exports.generateFieldWithHookFunc = (build, innerTable, joinCallback, isUnique = true, isConnection = false) => {
    const { pgSql: sql, graphql: { GraphQLNonNull, GraphQLList }, pgGetGqlTypeByTypeIdAndModifier, pgQueryFromResolveData: queryFromResolveData, getSafeAliasFromAlias, getSafeAliasFromResolveInfo, inflection, getTypeByName, } = build;
    const innerTableType = pgGetGqlTypeByTypeIdAndModifier(innerTable.type.id, null);
    const innerTableConnectionType = getTypeByName(inflection.connection(innerTable.name));
    let returnType = innerTableType;
    let fieldType = innerTableType;
    if (!isUnique) {
        if (isConnection) {
            returnType = innerTableConnectionType;
            fieldType = new GraphQLNonNull(innerTableConnectionType);
        }
        else {
            // using simple connection
            fieldType = new GraphQLList(new GraphQLNonNull(innerTableConnectionType));
        }
    }
    const rightTableTypeName = inflection.tableType(innerTable);
    return ({ getDataFromParsedResolveInfoFragment, addDataGenerator }) => {
        addDataGenerator((parsedResolveInfoFragment) => {
            return {
                pgQuery: (queryBuilder) => {
                    queryBuilder.select(() => {
                        const resolveData = getDataFromParsedResolveInfoFragment(parsedResolveInfoFragment, returnType);
                        const queryOptions = isUnique
                            ? {
                                useAsterisk: false,
                                asJson: true,
                                addNullCase: true,
                                withPagination: false,
                            }
                            : {
                                useAsterisk: innerTable.canUseAsterisk,
                                withPagination: isConnection,
                                withPaginationAsFields: false,
                                asJsonAggregate: !isConnection,
                            };
                        const innerTableAlias = sql.identifier(Symbol());
                        const query = queryFromResolveData(sql.identifier(innerTable.namespace.name, innerTable.name), innerTableAlias, resolveData, queryOptions, (innerQueryBuilder) => {
                            innerQueryBuilder.parentQueryBuilder = queryBuilder;
                            joinCallback(queryBuilder, innerQueryBuilder);
                        });
                        return sql.fragment `(${query})`;
                    }, getSafeAliasFromAlias(parsedResolveInfoFragment.alias));
                },
            };
        });
        return {
            description: `Reads through a \`${rightTableTypeName}\`.`,
            // type: new GraphQLNonNull(RightTableType),
            // This maybe should be nullable? because polymorphic foreign key
            // is not constraint
            type: fieldType,
            args: {},
            resolve: (data, _args, _context, resolveInfo) => {
                const safeAlias = getSafeAliasFromResolveInfo(resolveInfo);
                // return null;
                return data[safeAlias] || data[resolveInfo.fieldName];
            },
        };
    };
};
exports.getPrimaryKeys = (table) => {
    const foreignPrimaryConstraint = table.primaryKeyConstraint;
    if (!foreignPrimaryConstraint)
        return [];
    return foreignPrimaryConstraint.keyAttributes;
};
exports.getPrimaryKey = (table) => {
    const keys = exports.getPrimaryKeys(table);
    return keys[0];
};
exports.polyForeignKeyUnique = (build, foreignTable, c) => {
    const { pgIntrospectionResultsByKind: { constraint }, } = build;
    const { sourceTableId, sourceTableType } = exports.getSourceColumns(c);
    const isForeignKeyUnique = constraint.find((c) => {
        if (c.classId !== foreignTable.id)
            return false;
        // Only if the xxx_type, xxx_id are unique constraint
        if (c.keyAttributeNums.length !== 2)
            return false;
        // It must be an unique constraint
        if (c.type !== 'u')
            return false;
        // the two attributes must be xx_type, xx_id
        if (!c.keyAttributes.find(a => a.name === sourceTableId))
            return false;
        if (!c.keyAttributes.find(a => a.name === sourceTableType))
            return false;
        return true;
    });
    return !!isForeignKeyUnique;
};
exports.polySqlKeyMatch = (build, polyAlias, foreignAlias, fPKey, modelName, polyConstraint) => {
    const { pgSql: sql } = build;
    const { sourceTableId, sourceTableType } = exports.getSourceColumns(polyConstraint);
    return sql.query `(${sql.fragment `${polyAlias}.${sql.identifier(sourceTableId)} = ${foreignAlias}.${sql.identifier(fPKey.name)}`}) and (
${sql.fragment `${polyAlias}.${sql.identifier(sourceTableType)} = ${sql.value(modelName)}`})`;
};
exports.polymorphicCondition = (build, c, polyQueryBuilder, targetQueryBuilder, targetModelName, pKey) => {
    const targetTableAlias = targetQueryBuilder.getTableAlias();
    const polyTableAlias = polyQueryBuilder.getTableAlias();
    return exports.polySqlKeyMatch(build, polyTableAlias, targetTableAlias, pKey, targetModelName, c);
};
exports.ensureBuilderUniqueOrder = (build, innerBuilder, table) => {
    innerBuilder.beforeLock('orderBy', () => {
        // append order by primary key to the list of orders
        const { pgSql: sql } = build;
        if (!innerBuilder.isOrderUnique(false)) {
            innerBuilder.data.cursorPrefix = ['primary_key_asc'];
            if (table.primaryKeyConstraint) {
                const fPrimaryKeys = table.primaryKeyConstraint.keyAttributes;
                fPrimaryKeys.forEach((key) => {
                    innerBuilder.orderBy(sql.fragment `${innerBuilder.getTableAlias()}.
          ${sql.identifier(key.name)}`, true);
                });
            }
        }
        innerBuilder.setOrderIsUnique();
    });
};
//# sourceMappingURL=utils.js.map